@@ -7,6 +7,7 @@ from typing import (
     Iterator,
     List,
     Optional,
+    Tuple,
     Type,
     TypeVar,
     Union,
@@ -96,7 +97,6 @@ class ReverseRelation(Generic[MODEL]):
     async def __aiter__(self) -> AsyncGenerator[Any, MODEL]:
         if not self._fetched:
             self._set_result_for_query(await self)
-
         for val in self.related_objects:
             yield val
 
@@ -165,74 +165,42 @@ class ManyToManyRelation(ReverseRelation[MODEL]):
             return
         if not self.instance._saved_in_db:
             raise OperationalError(f"You should first call .save() on {self.instance}")
-        db = using_db if using_db else self.remote_model._meta.db
+        db = using_db or self.remote_model._meta.db
         pk_formatting_func = type(self.instance)._meta.pk.to_db_value
         related_pk_formatting_func = type(instances[0])._meta.pk.to_db_value
-        through_table = Table(self.field.through)
-        select_query = (
-            db.query_class.from_(through_table)
-            .where(
-                through_table[self.field.backward_key]
-                == pk_formatting_func(self.instance.pk, self.instance)
-            )
-            .select(self.field.backward_key, self.field.forward_key)
-        )
-        query = db.query_class.into(through_table).columns(
-            through_table[self.field.forward_key],
-            through_table[self.field.backward_key],
-        )
-
-        if len(instances) == 1:
-            criterion = through_table[self.field.forward_key] == related_pk_formatting_func(
-                instances[0].pk, instances[0]
-            )
-        else:
-            criterion = through_table[self.field.forward_key].isin(
-                [related_pk_formatting_func(i.pk, i) for i in instances]
-            )
-
-        select_query = select_query.where(criterion)
-
-        # TODO: This is highly inefficient. Should use UNIQUE index by default.
-        #  And optionally allow duplicates.
-        _, already_existing_relations_raw = await db.execute_query(str(select_query))
-        already_existing_relations = {
-            (
-                pk_formatting_func(r[self.field.backward_key], self.instance),
-                related_pk_formatting_func(r[self.field.forward_key], self.instance),
-            )
-            for r in already_existing_relations_raw
-        }
-
-        insert_is_required = False
+        pk_b = pk_formatting_func(self.instance.pk, self.instance)
+        pks_f: list = []
         for instance_to_add in instances:
             if not instance_to_add._saved_in_db:
                 raise OperationalError(f"You should first call .save() on {instance_to_add}")
             pk_f = related_pk_formatting_func(instance_to_add.pk, instance_to_add)
-            pk_b = pk_formatting_func(self.instance.pk, self.instance)
-            if (pk_b, pk_f) in already_existing_relations:
-                continue
-            query = query.insert(pk_f, pk_b)
-            insert_is_required = True
-        if insert_is_required:
+            pks_f.append(pk_f)
+        through_table = Table(self.field.through)
+        backward_key, forward_key = self.field.backward_key, self.field.forward_key
+        backward_field, forward_field = through_table[backward_key], through_table[forward_key]
+        select_query = (
+            db.query_class.from_(through_table).where(backward_field == pk_b).select(forward_key)
+        )
+        criterion = forward_field == pks_f[0] if len(pks_f) == 1 else forward_field.isin(pks_f)
+        select_query = select_query.where(criterion)
+
+        _, already_existing_relations_raw = await db.execute_query(str(select_query))
+        already_existing_forward_pks = {
+            related_pk_formatting_func(r[forward_key], self.instance)
+            for r in already_existing_relations_raw
+        }
+
+        if pks_f_to_insert := set(pks_f) - already_existing_forward_pks:
+            query = db.query_class.into(through_table).columns(forward_field, backward_field)
+            for pk_f in pks_f_to_insert:
+                query = query.insert(pk_f, pk_b)
             await db.execute_query(str(query))
 
     async def clear(self, using_db: "Optional[BaseDBAsyncClient]" = None) -> None:
         """
         Clears ALL relations.
         """
-        db = using_db if using_db else self.remote_model._meta.db
-        through_table = Table(self.field.through)
-        pk_formatting_func = type(self.instance)._meta.pk.to_db_value
-        query = (
-            db.query_class.from_(through_table)
-            .where(
-                through_table[self.field.backward_key]
-                == pk_formatting_func(self.instance.pk, self.instance)
-            )
-            .delete()
-        )
-        await db.execute_query(str(query))
+        await self._remove_or_clear(using_db=using_db)
 
     async def remove(
         self, *instances: MODEL, using_db: "Optional[BaseDBAsyncClient]" = None
@@ -242,30 +210,32 @@ class ManyToManyRelation(ReverseRelation[MODEL]):
 
         :raises OperationalError: remove() was called with no instances.
         """
-        db = using_db if using_db else self.remote_model._meta.db
         if not instances:
             raise OperationalError("remove() called on no instances")
+        await self._remove_or_clear(instances, using_db)
+
+    async def _remove_or_clear(
+        self,
+        instances: Optional[Tuple[MODEL, ...]] = None,
+        using_db: "Optional[BaseDBAsyncClient]" = None,
+    ) -> None:
+        db = using_db or self.remote_model._meta.db
         through_table = Table(self.field.through)
         pk_formatting_func = type(self.instance)._meta.pk.to_db_value
-        related_pk_formatting_func = type(instances[0])._meta.pk.to_db_value
 
-        if len(instances) == 1:
-            condition = (
-                through_table[self.field.forward_key]
-                == related_pk_formatting_func(instances[0].pk, instances[0])
-            ) & (
-                through_table[self.field.backward_key]
-                == pk_formatting_func(self.instance.pk, self.instance)
-            )
-        else:
-            condition = (
-                through_table[self.field.backward_key]
-                == pk_formatting_func(self.instance.pk, self.instance)
-            ) & (
-                through_table[self.field.forward_key].isin(
+        condition = through_table[self.field.backward_key] == pk_formatting_func(
+            self.instance.pk, self.instance
+        )
+        if instances:
+            related_pk_formatting_func = type(instances[0])._meta.pk.to_db_value
+            if len(instances) == 1:
+                condition &= through_table[self.field.forward_key] == related_pk_formatting_func(
+                    instances[0].pk, instances[0]
+                )
+            else:
+                condition &= through_table[self.field.forward_key].isin(
                     [related_pk_formatting_func(i.pk, i) for i in instances]
                 )
-            )
         query = db.query_class.from_(through_table).where(condition).delete()
         await db.execute_query(str(query))
 
@@ -289,20 +259,16 @@ class RelationalField(Field[MODEL]):
     if TYPE_CHECKING:
 
         @overload
-        def __get__(self, instance: None, owner: Type["Model"]) -> "RelationalField[MODEL]":
-            ...
+        def __get__(self, instance: None, owner: Type["Model"]) -> "RelationalField[MODEL]": ...
 
         @overload
-        def __get__(self, instance: "Model", owner: Type["Model"]) -> MODEL:
-            ...
+        def __get__(self, instance: "Model", owner: Type["Model"]) -> MODEL: ...
 
         def __get__(
             self, instance: Optional["Model"], owner: Type["Model"]
-        ) -> "RelationalField[MODEL] | MODEL":
-            ...
+        ) -> "RelationalField[MODEL] | MODEL": ...
 
-        def __set__(self, instance: "Model", value: MODEL) -> None:
-            ...
+        def __set__(self, instance: "Model", value: MODEL) -> None: ...
 
     def describe(self, serializable: bool) -> dict:
         desc = super().describe(serializable)
@@ -310,6 +276,12 @@ class RelationalField(Field[MODEL]):
         del desc["db_column"]
         return desc
 
+    @classmethod
+    def validate_model_name(cls, model_name: str) -> None:
+        if len(model_name.split(".")) != 2:
+            field_type = cls.__name__.replace("Instance", "")
+            raise ConfigurationError(f'{field_type} accepts model name in format "app.Model"')
+
 
 class ForeignKeyFieldInstance(RelationalField[MODEL]):
     def __init__(
@@ -320,8 +292,7 @@ class ForeignKeyFieldInstance(RelationalField[MODEL]):
         **kwargs: Any,
     ) -> None:
         super().__init__(None, **kwargs)  # type: ignore
-        if len(model_name.split(".")) != 2:
-            raise ConfigurationError('Foreign key accepts model name in format "app.Model"')
+        self.validate_model_name(model_name)
         self.model_name = model_name
         self.related_name = related_name
         if on_delete not in set(OnDelete):
@@ -363,8 +334,7 @@ class OneToOneFieldInstance(ForeignKeyFieldInstance[MODEL]):
         on_delete: OnDelete = CASCADE,
         **kwargs: Any,
     ) -> None:
-        if len(model_name.split(".")) != 2:
-            raise ConfigurationError('OneToOneField accepts model name in format "app.Model"')
+        self.validate_model_name(model_name)
         super().__init__(model_name, related_name, on_delete, unique=True, **kwargs)
 
 
@@ -384,13 +354,13 @@ class ManyToManyFieldInstance(RelationalField[MODEL]):
         related_name: str = "",
         on_delete: OnDelete = CASCADE,
         field_type: "Type[MODEL]" = None,  # type: ignore
+        create_unique_index: bool = True,
         **kwargs: Any,
     ) -> None:
         # TODO: rename through to through_table
         # TODO: add through to use a Model
         super().__init__(field_type, **kwargs)
-        if len(model_name.split(".")) != 2:
-            raise ConfigurationError('Foreign key accepts model name in format "app.Model"')
+        self.validate_model_name(model_name)
         self.model_name: str = model_name
         self.related_name: str = related_name
         self.forward_key: str = forward_key or f"{model_name.split('.')[1].lower()}_id"
@@ -398,6 +368,7 @@ class ManyToManyFieldInstance(RelationalField[MODEL]):
         self.through: str = through  # type: ignore
         self._generated: bool = False
         self.on_delete = on_delete
+        self.create_unique_index = create_unique_index
 
     def describe(self, serializable: bool) -> dict:
         desc = super().describe(serializable)
@@ -420,8 +391,7 @@ def OneToOneField(
     *,
     null: Literal[True],
     **kwargs: Any,
-) -> "OneToOneNullableRelation[MODEL]":
-    ...
+) -> "OneToOneNullableRelation[MODEL]": ...
 
 
 @overload
@@ -432,8 +402,7 @@ def OneToOneField(
     db_constraint: bool = True,
     null: Literal[False] = False,
     **kwargs: Any,
-) -> "OneToOneRelation[MODEL]":
-    ...
+) -> "OneToOneRelation[MODEL]": ...
 
 
 def OneToOneField(
@@ -497,8 +466,7 @@ def ForeignKeyField(
     *,
     null: Literal[True],
     **kwargs: Any,
-) -> "ForeignKeyNullableRelation[MODEL]":
-    ...
+) -> "ForeignKeyNullableRelation[MODEL]": ...
 
 
 @overload
@@ -509,8 +477,7 @@ def ForeignKeyField(
     db_constraint: bool = True,
     null: Literal[False] = False,
     **kwargs: Any,
-) -> "ForeignKeyRelation[MODEL]":
-    ...
+) -> "ForeignKeyRelation[MODEL]": ...
 
 
 def ForeignKeyField(
@@ -573,6 +540,7 @@ def ManyToManyField(
     related_name: str = "",
     on_delete: OnDelete = CASCADE,
     db_constraint: bool = True,
+    create_unique_index: bool = True,
     **kwargs: Any,
 ) -> "ManyToManyRelation[Any]":
     """
@@ -618,6 +586,9 @@ def ManyToManyField(
                 Can only be set is field has a ``default`` set.
             ``field.NO_ACTION``:
                 Take no action.
+    ``create_unique_index``:
+        Controls whether or not a unique index should be created in the database to speed up select queries.
+        The default is True. If you want to allow repeat records, set this to False.
     """
 
     return ManyToManyFieldInstance(  # type: ignore
@@ -628,6 +599,7 @@ def ManyToManyField(
         related_name,
         on_delete=on_delete,
         db_constraint=db_constraint,
+        create_unique_index=create_unique_index,
         **kwargs,
     )
 
